# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from base64 import b64encode, b64decode
from functools import partial

from markupsafe import escape

from fresco import core, context


__version__ = "1.0.0"


class FlashMiddleware(object):
    """
    Middleware to facilitate setting and displaying messages to the user across
    an http redirect.

    Usage::

        from fresco import context, flash

        def app1():
            flash.info("Operation completed")
            flash.warn("1 warning encountered")
            return Response.redirect('/app2.html')

        app1 = FlashMiddleware(app1)

        def app2():
            return Response("<p>Message was: %s</p>" %
                            context.request.environ['fresco.flash_messages'])

    """

    INFO = "INFO"
    WARN = "WARN"
    ERROR = "ERROR"

    __name__ = "FlashMiddleware"

    #: Key used to reference the middleware in the WSGI environ
    env_key = "fresco.flash"

    #: Key used for stash of messages in the WSGI environ
    env_key_messages = "fresco.flash.messages"

    #: Name for messages stored in cookies
    cookie_key = "_flash"

    def __init__(self, app):
        self.app = app
        try:
            self.request_class = core.request_class
        except AttributeError:
            self.request_class = None

    def _flash(self, messages, message, category=INFO):
        messages.append((category, escape(message)))

    def _start_response(
        self, environ, start_response, status, headers, exc_info=None
    ):
        messages = environ[self.env_key_messages]
        script_name = environ.get("SCRIPT_NAME", "") or "/"

        # Redirect response: set flash cookie so the messages are persisted to
        # the next request
        if status[:2] == "30" and messages:
            cookie = ",".join(
                b64encode(("%s:%s" % (c, m)).encode("utf8")).decode("ascii")
                for c, m in messages
            )
            headers.append(
                (
                    "Set-Cookie",
                    "%s=%s;path=%s" % (self.cookie_key, cookie, script_name,),
                )
            )

        # Clear the cookie so that messages are not repeated on subsequent
        # requests
        elif self.cookie_key in environ.get("HTTP_COOKIE", ""):
            headers.append(
                (
                    "Set-Cookie",
                    "%s=;Expires=Tue, 01 Jan 1980 00:00:00;Path=%s"
                    % (self.cookie_key, script_name),
                )
            )

        return start_response(status, headers, exc_info)

    def __call__(self, environ, start_response):

        environ[self.env_key_messages] = messages = []
        environ[self.env_key] = partial(self._flash, messages)

        if self.cookie_key in environ.get("HTTP_COOKIE", ""):
            if self.request_class is None:
                self.request_class = context.app.request_class
            request = self.request_class(environ)
            cookie = request.cookies.get(self.cookie_key, None)
            if cookie:
                messages.extend(
                    b64decode(item.encode("ascii")).decode("utf8").split(":", 1)
                    for item in cookie.split(",")
                )

        return self.app(
            environ, partial(self._start_response, environ, start_response)
        )



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps
from markupsafe import Markup, escape

from fresco import FrescoApp, context, GET, Response
from fresco_flash import FlashMiddleware, flash


def flash_messages():
    request = context.request
    for item in request.query.getlist("m"):
        flash.info(item)
    if "r" in request.query:
        return Response.redirect(request.query.get("r"))
    return Response([dumps(flash.messages())])


app = FrescoApp()
app.route("/flash", GET, flash_messages)
app.add_middleware(FlashMiddleware)


def doflash(*messages):
    with app.requestcontext():
        for m in messages:
            flash.info(m)
        return flash.messages()


def doflash_cross_request(*messages):
    def flashit_and_redirect(environ, start_response):
        for m in messages:
            flash.info(m)
        start_response("302 Found", [("location", "/")])
        return []

    def start_response(status, headers, exc_info=None):
        start_response.status = status
        start_response.headers = headers

    with app.requestcontext() as c:
        FlashMiddleware(flashit_and_redirect)(c.request.environ, start_response)
        cookie = dict(start_response.headers)["Set-Cookie"]
        cookie = cookie.split(";", 1)[0]

    with app.requestcontext(HTTP_COOKIE=cookie) as c:
        wsgiapp = FlashMiddleware(lambda *args, **kwargs: [])
        wsgiapp(c.request.environ, start_response)
        return flash.messages()


def doflash_one(message):
    return doflash(message)[0][1]


def doflash_one_cross_request(message):
    return doflash_cross_request(message)[0][1]


class TestFlash(object):
    def test_flash(self):
        assert doflash("hello") == [("INFO", Markup("hello"))]

    def test_flash_with_redirect(self):
        assert doflash_cross_request("hello") == [("INFO", Markup("hello"))]

    def test_multi_flash_with_redirect(self):
        assert doflash_cross_request("hello", "world") == [
            ("INFO", Markup("hello")),
            ("INFO", Markup("world")),
        ]

    def test_multi_flash_with_metachars_and_redirect(self):
        assert doflash_cross_request("hel:lo", "wo,r;ld") == [
            ("INFO", Markup("hel:lo")),
            ("INFO", Markup("wo,r;ld")),
        ]


class TestEscaping(object):
    def test_it_escapes_embedded_html(self):

        with app.requestcontext("/"):
            flash.info("2 > 1")
            assert flash.messages() == [("INFO", "2 &gt; 1")]

    def test_it_preserves_untainted_markup(self):

        with app.requestcontext("/"):
            flash.info(Markup("2 > 1"))
            assert flash.messages() == [("INFO", "2 > 1")]

    def test_it_preserves_tainting_info(self):

        msg = doflash_one_cross_request("2 > 1")
        assert str(escape(msg)) == "2 &gt; 1"

        msg = doflash_one_cross_request(Markup("2 > 1"))
        assert str(escape(msg)) == "2 > 1"
